from dataclasses import dataclass
import os
from typing import Optional, Set
from clddp.args.base import AutoRunNameArgumentsMixIn, DumpableArgumentsMixIn
from clddp.utils import parse_cli
from clddp.retriever import Separator, Pooling, SimilarityFunction
from clddp.dm import Split
from transformers import IntervalStrategy, TrainingArguments


@dataclass
class RetrievalTrainingArguments(
    AutoRunNameArgumentsMixIn, DumpableArgumentsMixIn, TrainingArguments
):
    project: str = "clddp"  # Used for wandb
    checkpoint_dir: str = "checkpoints"

    # Retriever-related:
    query_model_name_or_path: str = (
        "sentence-transformers/msmarco-distilbert-base-tas-b"
    )
    passage_model_name_or_path: Optional[str] = None
    query_prompt: Optional[str] = None
    passage_prompt: Optional[str] = None
    shared_encoder: bool = True
    sep: Separator = Separator.blank
    pooling: Pooling = Pooling.cls
    similarity_function: SimilarityFunction = SimilarityFunction.dot_product
    max_length: int = 350
    sim_scale: float = 1.0
    fp16: bool = True

    # Data-related:
    train_data: Optional[str] = None
    train_dataloader: Optional[str] = None
    dev_data: Optional[str] = None
    dev_dataloader: Optional[str] = None
    test_data: Optional[str] = None
    test_dataloader: Optional[str] = None

    # General training hyperparameters:
    do_train: bool = True
    per_device_train_batch_size: int = 64
    gradient_accumulation_steps: int = 1
    per_device_eval_batch_size: int = 32
    learning_rate: int = 2e-5
    num_train_epochs: int = 5
    warmup_ratio: float = 0.1
    warmup_steps: int = 0

    # Evaluation-related:
    do_dev: bool = True  # whether to evaluate on dev during training
    quick_dev: bool = False  # whether to keep only the labeled passages + sampled passages as the collection for quick dev
    do_test: bool = True
    quick_test: bool = False  # whether to keep only the labeled passages + sampled passages as the collection for quick test
    metric_for_best_model: str = "ndcg_cut_10"
    dev_split: Split = Split.dev  # Note that some datasets do not have a dev split
    test_split: Split = Split.test
    topk: int = 1000
    save_steps: float = 0.1
    eval_steps: float = 0.1  # dev steps
    evaluation_strategy: IntervalStrategy = IntervalStrategy.STEPS

    # Others:
    logging_steps: float = 1
    ddp_find_unused_parameters = False

    @property
    def escaped_args(self) -> Set[str]:
        """Escaped items for building the run_name."""
        return {
            "checkpoint_dir",
            "logging_steps",
            "ddp_find_unused_parameters",
            "save_steps",
            "train_dataloader",
            "dev_dataloader",
            "test_dataloader",
        }

    def __post_init__(self) -> None:
        os.environ[
            "WANDB_PROJECT"
        ] = self.project  # HF's trainer relis on this to know the project name
        self.dev_split = Split(self.dev_split)
        self.test_split = Split(self.test_split)
        self.sep = Separator(self.sep)
        self.pooling = Pooling(self.pooling)
        self.similarity_function = SimilarityFunction(self.similarity_function)
        super().__post_init__()

    def build_output_dir(self) -> str:
        return os.path.join(self.checkpoint_dir, self.run_name)


if __name__ == "__main__":
    print(
        parse_cli(RetrievalTrainingArguments).build_output_dir()
    )  # For creating the logging path



"""Data models."""

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
import json
import logging
import random
from typing import Callable, Dict, Iterable, List, Optional, Type

import tqdm


@dataclass
class Passage:
    passage_id: str
    text: str
    title: Optional[str] = None


@dataclass
class Query:
    query_id: str
    text: str


@dataclass
class JudgedPassage:
    query: Query
    passage: Passage
    judgement: int

    @staticmethod
    def build_qrels(judeged_passages: List[JudgedPassage]) -> Dict[str, Dict[str, int]]:
        """Build the qrels for trec_eval https://github.com/cvangysel/pytrec_eval."""
        qrels = {}
        for jp in judeged_passages:
            qrels.setdefault(jp.query.query_id, {})
            qrels[jp.query.query_id][jp.passage.passage_id] = jp.judgement
        return qrels

    @staticmethod
    def get_unique_queries(judged_passages: List[JudgedPassage]) -> List[Query]:
        qid2query = {jpsg.query.query_id: jpsg.query for jpsg in judged_passages}
        queries = list(qid2query.values())
        return queries


@dataclass
class ScoredPassageID:
    passage_id: str
    score: float


@dataclass
class RetrievedPassageIDList:
    query_id: str
    scored_passage_ids: List[ScoredPassageID]

    @staticmethod
    def build_trec_scores(
        retrieved: List[RetrievedPassageIDList],
    ) -> Dict[str, Dict[str, float]]:
        trec_scores = {
            rpidl.query_id: {
                spid.passage_id: spid.score for spid in rpidl.scored_passage_ids
            }
            for rpidl in retrieved
        }
        return trec_scores

    @classmethod
    def merge(
        cls: Type[RetrievedPassageIDList],
        rpidls: List[RetrievedPassageIDList],
        topk: Optional[int] = None,
    ) -> RetrievedPassageIDList:
        qids = list(set(rpidl.query_id for rpidl in rpidls))
        assert len(qids) == 1, "Can only merge results for the same query"
        pid2score = {
            spid.passage_id: spid.score
            for rpidl in rpidls
            for spid in rpidl.scored_passage_ids
        }
        if topk:
            pid2score = dict(
                sorted(
                    pid2score.items(), key=lambda pid_score: pid_score[1], reverse=True
                )[:topk]
            )
        return RetrievedPassageIDList(
            query_id=qids[0],
            scored_passage_ids=[
                ScoredPassageID(passage_id=pid, score=score)
                for pid, score in pid2score.items()
            ],
        )

    def dump_trec_csv(
        retrieval_results: List[RetrievedPassageIDList], fpath: str
    ) -> None:
        with open(fpath, "w") as f:
            for ranking in tqdm.tqdm(
                retrieval_results, desc="Saving retrieval results"
            ):
                sorted_spsg_ids = sorted(
                    ranking.scored_passage_ids,
                    key=lambda spsg_id: spsg_id.score,
                    reverse=True,
                )
                for i, spsg_id in enumerate(sorted_spsg_ids):
                    row = (
                        ranking.query_id,
                        "Q0",
                        spsg_id.passage_id,
                        str(i + 1),
                        str(spsg_id.score),
                        "clddp",
                    )
                    f.write(" ".join(row) + "\n")
        logging.info(f"Dumped retrieval results to {fpath} in TREC format.")


class Split(str, Enum):
    train = "train"
    dev = "dev"
    test = "test"


@dataclass
class RetrievalDataset:
    collection_iter_fn: Callable[
        [], Iterable[Passage]
    ]  # This design is for handling very large collections
    collection_size: int
    judged_passages_train: Optional[List[JudgedPassage]] = None
    judged_passages_dev: Optional[List[JudgedPassage]] = None
    judged_passages_test: Optional[List[JudgedPassage]] = None

    @property
    def collection_iter(self) -> Iterable[Passage]:
        return self.collection_iter_fn()

    def get_judged_passages(self, split: Split) -> Optional[List[JudgedPassage]]:
        return {
            Split.train: self.judged_passages_train,
            Split.dev: self.judged_passages_dev,
            Split.test: self.judged_passages_test,
        }[split]

    def to_quick_version(
        self,
        split: Split,
        seed: int = 42,
        progress_bar: bool = False,
        save_pids_to_fpath: Optional[str] = None,
    ) -> RetrievalDataset:
        """Build a quick version for the dataset with a trimmed collection containing only the judged passages and a few sampled passages."""
        judged_passages = self.get_judged_passages(split)
        random_state = random.Random(seed)
        jpsgs_ids = {jpsg.passage.passage_id for jpsg in judged_passages}
        sampled_indices = set(
            random_state.sample(list(range(self.collection_size)), len(jpsgs_ids))
        )
        sampled_passages = []
        for i, psg in enumerate(
            tqdm.tqdm(
                self.collection_iter,
                total=self.collection_size,
                desc=f"Building quick version of the {split} dataset",
                disable=not progress_bar,
            )
        ):
            if i in sampled_indices:
                sampled_passages.append(psg)
        quick_collection = [jpsg.passage for jpsg in judged_passages] + sampled_passages
        pid2psg = {psg.passage_id: psg for psg in quick_collection}
        dedup_quick_collection = list(pid2psg.values())
        random_state.shuffle(dedup_quick_collection)
        quick_dataset = RetrievalDataset(
            collection_iter_fn=lambda: iter(dedup_quick_collection),
            collection_size=len(dedup_quick_collection),
        )
        # Only the specified split is kept:
        if split is Split.train:
            quick_dataset.judged_passages_train = judged_passages
        elif split is Split.dev:
            quick_dataset.judged_passages_dev = judged_passages
        else:
            assert split is Split.test
            quick_dataset.judged_passages_test = judged_passages
        if save_pids_to_fpath:
            pids = [psg.passage_id for psg in quick_dataset.collection_iter]
            with open(save_pids_to_fpath, "w") as f:
                json.dump(pids, f)
            logging.info(
                f"Saved passage IDs of the quick {split} dataset to {save_pids_to_fpath} (num.: {len(pids)})"
            )
        return quick_dataset



from enum import Enum
import json
import logging
import os
from typing import Dict, List, Tuple
import numpy as np
import pytrec_eval
from clddp.dm import (
    JudgedPassage,
    RetrievalDataset,
    RetrievedPassageIDList,
    Split,
)
from clddp.dataloader import load_dataset
from clddp.retriever import Retriever
from clddp.search import search
import torch.distributed as dist
from clddp.utils import (
    get_rank,
    is_device_zero,
    set_logger_format,
)
from clddp.args.evaluation import EvaluationArguments
from transformers import HfArgumentParser


class RetrievalMetric(str, Enum):
    map_string = "map_cut"
    ndcg_string = "ndcg_cut"
    recall_string = "recall"
    precision_string = "P"
    rr_string = "recip_rank"

    @staticmethod
    def cutoff(metric_string: str) -> int:
        return int(metric_string.split("_")[-1])

    def at(self, k: int) -> str:
        return f"{self}_{k}"

    def trec_string(self, k_values: Tuple[int]) -> str:
        if self is RetrievalMetric.rr_string:
            return self

        return f"{self}." + ",".join([str(k) for k in k_values])


class RetrievalEvaluator:
    def __init__(
        self,
        eval_dataset: RetrievalDataset,
        split: Split,
        metrics: Tuple[str] = (
            RetrievalMetric.ndcg_string.at(10),
            RetrievalMetric.rr_string.at(10),
            RetrievalMetric.recall_string.at(100),
        ),
        precision: int = 4,
    ) -> None:
        self.judged_passages = eval_dataset.get_judged_passages(split)
        self.queries = JudgedPassage.get_unique_queries(self.judged_passages)
        self.qrels = JudgedPassage.build_qrels(self.judged_passages)
        self.metrics = metrics
        self.precision = precision

    def __call__(
        self, retrieved_passage_id_lists: List[RetrievedPassageIDList]
    ) -> Dict[str, float]:
        trec_scores = RetrievedPassageIDList.build_trec_scores(
            retrieved_passage_id_lists
        )
        eval_scores = self._pytrec_eval(trec_scores=trec_scores, qrels=self.qrels)
        report = self._build_report(eval_scores=eval_scores)
        return report

    def _post_process_rr(self, eval_scores: Dict[str, Dict[str, float]]) -> None:
        """pytrec_eval does not support MRR at K originally."""
        rr_cutoffs = [
            int(m.split("_")[-1])
            for m in self.metrics
            if RetrievalMetric.rr_string in m
        ]
        for k in rr_cutoffs:
            min_rr = 1 / k
            for metric2score in eval_scores.values():
                score = metric2score[RetrievalMetric.rr_string]
                if score < min_rr:
                    score = 0
                metric2score[RetrievalMetric.rr_string.at(k)] = score

    def _pytrec_eval(
        self, trec_scores: Dict[str, Dict[str, float]], qrels: Dict[str, Dict[str, int]]
    ) -> Dict[str, Dict[str, float]]:
        logging.info("Running pytrec-eval")
        evaluator = pytrec_eval.RelevanceEvaluator(qrels, self.metrics)
        eval_scores: Dict[str, Dict[str, float]] = evaluator.evaluate(trec_scores)
        self._post_process_rr(eval_scores)
        return eval_scores

    def _build_report(
        self, eval_scores: Dict[str, Dict[str, float]], prefix: str = ""
    ) -> Dict[str, float]:
        report: Dict[str, float] = {}
        for metric in self.metrics:
            item_name = f"{prefix}{metric}"
            report[item_name] = round(
                np.mean([m2s[metric] for m2s in eval_scores.values()]).tolist(),
                self.precision,
            )
        report = dict(sorted(report.items(), key=lambda kv: kv[0]))
        return report


def search_and_evaluate(
    retriever: Retriever,
    eval_dataset: RetrievalDataset,
    split: Split,
    topk: int,
    per_device_eval_batch_size: int,
    fp16: bool,
    metric_key_prefix: str = "eval",
) -> Tuple[List[RetrievedPassageIDList], Dict[str, float]]:
    # Build evaluator and do search:
    evaluator = RetrievalEvaluator(eval_dataset=eval_dataset, split=split)
    rpidls = search(
        retriever=retriever,
        collection_iter=eval_dataset.collection_iter,
        collection_size=eval_dataset.collection_size,
        queries=evaluator.queries,
        topk=topk,
        per_device_eval_batch_size=per_device_eval_batch_size,
        fp16=fp16,
    )

    # Calculate scores:
    report = evaluator(rpidls)
    report_prefixed = {}
    for k, v in report.items():
        report_prefixed[f"{metric_key_prefix}_{k}"] = v  # Required by HF's trainer

    if dist.is_initialized():
        dist.barrier()

    logging.info(f"Evaluation results: {report_prefixed}")
    return rpidls, report_prefixed


if __name__ == "__main__":
    """Example script: torchrun --nproc_per_node=4 --master_port=29501 -m clddp.evaluation --checkpoint_dir=..."""
    from clddp.utils import parse_cli, initialize_ddp

    set_logger_format(logging.INFO if is_device_zero() else logging.WARNING)
    initialize_ddp()
    args = parse_cli(EvaluationArguments)
    if is_device_zero():
        args.dump_arguments()

    retriever = Retriever.from_pretrained(args.checkpoint_dir)
    if args.query_prompt is not None:
        retriever.set_query_prompt(args.query_prompt)
    if args.passage_prompt is not None:
        retriever.set_passage_prompt(args.passage_prompt)
    retriever.set_device(get_rank())
    eval_dataset = load_dataset(
        enable=True, dataloader_name=args.dataloader, data_name_or_path=args.data_dir
    )
    retrieval_results, report = search_and_evaluate(
        retriever=retriever,
        eval_dataset=eval_dataset,
        split=args.split,
        topk=args.topk,
        per_device_eval_batch_size=args.per_device_eval_batch_size,
        fp16=args.fp16,
    )
    fretrieved = os.path.join(args.output_dir, "retrieval_results.txt")
    if is_device_zero():
        RetrievedPassageIDList.dump_trec_csv(
            retrieval_results=retrieval_results, fpath=fretrieved
        )
        freport = os.path.join(args.output_dir, "metrics.json")
        with open(freport, "w") as f:
            json.dump(report, f, indent=4)
        logging.info(f"Saved evaluation metrics to {freport}.")
    logging.info("Done")



from abc import ABC, abstractstaticmethod
import csv
import logging
import os
from typing import Dict, List, Optional
from clddp.dm import JudgedPassage, Passage, Query, RetrievalDataset, Split
from clddp.utils import is_device_zero, tqdm_ropen
import ujson


class BaseDataLoader(ABC):
    @abstractstaticmethod
    def load_data(data_name_or_path: str, progress_bar: bool) -> RetrievalDataset:
        pass


class BEIRDataloader(BaseDataLoader):
    @staticmethod
    def load_qrels_from_beir(
        data_dir: str, split: Split
    ) -> Optional[Dict[str, Dict[str, int]]]:
        qrel_path = os.path.join(data_dir, "qrels", f"{split}.tsv")
        if not os.path.exists(qrel_path):
            logging.info(f"Found no {split} split under {data_dir}: {qrel_path}")
            return None

        reader = csv.reader(
            open(qrel_path, encoding="utf-8"),
            delimiter="\t",
            quoting=csv.QUOTE_MINIMAL,
        )
        next(reader)  # skip the header

        qrels = {}
        for row in reader:
            query_id, corpus_id, score = row[0], row[1], int(row[2])
            qrels.setdefault(query_id, {})
            qrels[query_id][corpus_id] = score

        return qrels

    @staticmethod
    def load_data(data_name_or_path: str, progress_bar: bool) -> RetrievalDataset:
        # Loading collection:
        collection: List[Passage] = []
        collection_path = os.path.join(data_name_or_path, "corpus.jsonl")
        for line in tqdm_ropen(
            fpath=collection_path,
            desc=f"Loading collection from {collection_path}",
            pbar=progress_bar,
        ):
            psg_json = ujson.loads(line)
            passage = Passage(passage_id=psg_json["_id"], text=psg_json["text"])
            if psg_json["title"]:
                passage.title = psg_json["title"]
            collection.append(passage)
        pid2psg = {psg.passage_id: psg for psg in collection}

        # Loading queries:
        queries: List[Query] = []
        queries_path = os.path.join(data_name_or_path, "queries.jsonl")
        for line in tqdm_ropen(
            fpath=queries_path,
            desc=f"Loading queries from {queries_path}",
            pbar=progress_bar,
        ):
            query_json = ujson.loads(line)
            queries.append(Query(query_id=query_json["_id"], text=query_json["text"]))
        qid2query = {query.query_id: query for query in queries}

        # Loading qrels:
        split = Split.train
        jpsgs_train = []
        jpsgs_dev = []
        jpsgs_test = []
        for jpsgs, split in zip([jpsgs_train, jpsgs_dev, jpsgs_test], Split):
            qrels = BEIRDataloader.load_qrels_from_beir(
                data_dir=data_name_or_path, split=split
            )
            if qrels is None:
                continue

            for qid, rels in qrels.items():
                for pid, rel in rels.items():
                    jpsg = JudgedPassage(
                        query=qid2query[qid], passage=pid2psg[pid], judgement=rel
                    )
                    jpsgs.append(jpsg)

        # Return loaded data:
        return RetrievalDataset(
            collection_iter_fn=lambda: iter(collection),
            collection_size=len(collection),
            judged_passages_train=jpsgs_train,
            judged_passages_dev=jpsgs_dev,
            judged_passages_test=jpsgs_test,
        )


DATA_LOADER_LOOKUP: Dict[str, BaseDataLoader] = {"beir": BEIRDataloader}


def load_dataset(
    enable: bool, dataloader_name: Optional[str], data_name_or_path: Optional[str]
) -> Optional[RetrievalDataset]:
    if enable:
        assert dataloader_name is not None
        assert data_name_or_path is not None
        assert dataloader_name in DATA_LOADER_LOOKUP
        return DATA_LOADER_LOOKUP[dataloader_name].load_data(
            data_name_or_path=data_name_or_path, progress_bar=is_device_zero()
        )
    else:
        return None



from dataclasses import dataclass
import os
from typing import Optional, Set
from clddp.args.base import AutoRunNameArgumentsMixIn, DumpableArgumentsMixIn
from clddp.dm import Split
from clddp.utils import parse_cli


@dataclass
class SearchArguments(AutoRunNameArgumentsMixIn, DumpableArgumentsMixIn):
    checkpoint_dir: Optional[str] = None
    data_dir: Optional[str] = None
    dataloader: Optional[str] = None
    search_dir: str = "search"
    split: Split = Split.test
    topk: int = 1000
    per_device_eval_batch_size: int = 32
    fp16: bool = True
    query_prompt: Optional[str] = None
    passage_prompt: Optional[str] = None

    def __post_init__(self) -> None:
        self.split = Split(self.split)
        return super().__post_init__()

    @property
    def escaped_args(self) -> Set[str]:
        return {"search_dir", "dataloader"}

    def build_output_dir(self) -> str:
        return os.path.join(self.search_dir, self.run_name)


if __name__ == "__main__":
    print(
        parse_cli(SearchArguments).build_output_dir()
    )  # For creating the logging path



from __future__ import annotations
from dataclasses import asdict, dataclass
import dataclasses
from enum import Enum
from functools import partial
from itertools import chain
import json
import logging
import os
from typing import Any, Dict, Iterator, List, Optional, Type, Union
import torch
from sentence_transformers.util import (
    cos_sim,
    dot_score,
    pairwise_cos_sim,
    pairwise_dot_score,
)
from transformers import (
    AutoModelForTextEncoding,
    AutoTokenizer,
    PreTrainedModel,
    BatchEncoding,
)
from transformers.models.auto.modeling_auto import MODEL_FOR_TEXT_ENCODING_MAPPING_NAMES
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
import torch.distributed as dist
from clddp.dm import Passage, Query
from clddp.utils import dist_gather_tensor, colbert_score

MODEL_FOR_TEXT_ENCODING_MAPPING_NAMES[
    "mpnet"
] = "MPNetModel"  # Missing item in the current transformers version


def pairwise_maxsim(
    qembs: torch.Tensor, pembs: torch.Tensor, pmask: torch.BoolTensor
) -> torch.Tensor:
    assert pmask is not None
    return colbert_score(Q=qembs, D_padded=pembs, D_mask=pmask)


def maxsim(
    qembs: torch.Tensor,
    cembs: torch.Tensor,
    cmask: torch.BoolTensor,
    batch_size_query: int = 1024,
    batch_size_passage: int = 16,
) -> torch.Tensor:
    assert cmask is not None
    scores = []
    for bq in range(0, len(qembs), batch_size_query):
        eq = bq + batch_size_query
        qbatch = qembs[bq:eq]
        scores_qbatch = []
        for bc in range(0, len(cembs), batch_size_passage):
            ec = bc + batch_size_passage
            scores_4d = cembs[None, bc:ec] @ qbatch[:, None].to(
                dtype=cembs.dtype
            ).transpose(
                2, 3
            )  # (nqueries, npassages, passage_length, query_length)
            scores_4d += (
                ~cmask[None, bc:ec] * -9999
            )  # (npassages, passage_length, 1) -> (1, npassages, passage_length, 1)
            scores_4d_max: torch.Tensor = scores_4d.max(dim=2)[
                0
            ]  # (nqueries, npassages, query_length)
            scores_2d = scores_4d_max.sum(dim=-1)  # (nqueries, npassages)
            scores_qbatch.append(scores_2d)
        scores.append(torch.cat(scores_qbatch, dim=-1))
    return torch.cat(scores, dim=0)


class Pooling(str, Enum):
    cls = "cls"
    mean = "mean"
    splade = "splade"
    sum = "sum"
    max = "max"
    no_pooling = "no_pooling"

    def __call__(
        self, token_embeddings: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        """Pooling: (bsz, seq_len, hdim) -> (bsz, hdim) or return the input without pooling."""

        if self == Pooling.cls:
            return token_embeddings[:, 0:1].sum(dim=1)
        elif self == Pooling.mean:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
                input_mask_expanded.sum(1), min=1e-9
            )
        elif self == Pooling.sum:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            return torch.sum(token_embeddings * input_mask_expanded, 1)
        elif self == Pooling.max:
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            token_embeddings[
                input_mask_expanded == 0
            ] = -1e9  # Set padding tokens to large negative value
            return torch.max(token_embeddings, 1)[0]
        elif self == Pooling.splade:
            pooled: torch.Tensor = getattr(
                torch.max(
                    torch.log(1 + torch.relu(token_embeddings))
                    * attention_mask.unsqueeze(-1),
                    dim=1,
                ),
                "values",
            )
            return pooled
        elif self == Pooling.no_pooling:
            return token_embeddings
        else:
            return NotImplementedError


class SimilarityFunction(str, Enum):
    """Vector distance between embeddings."""

    dot_product = "dot_product"
    cos_sim = "cos_sim"
    maxsim = "maxsim"

    def __call__(
        self,
        query_embeddings: torch.Tensor,
        passage_embeddings: torch.Tensor,
        passage_mask: Optional[torch.BoolTensor] = None,
        pairwise: bool = False,
    ) -> torch.Tensor:
        """Run the score function over query and passage embeddings."""
        if pairwise:
            # return shape: (npairs,)
            assert len(query_embeddings) == len(passage_embeddings)
            fn = {
                SimilarityFunction.dot_product: pairwise_dot_score,
                SimilarityFunction.cos_sim: pairwise_cos_sim,
                SimilarityFunction.maxsim: partial(pairwise_maxsim, pmask=passage_mask),
            }[self]
        else:
            # resturn shape: (nqueries, npassages)
            fn = {
                SimilarityFunction.dot_product: dot_score,
                SimilarityFunction.cos_sim: cos_sim,
                SimilarityFunction.maxsim: partial(maxsim, pmask=passage_mask),
            }[self]
        scores = fn(query_embeddings, passage_embeddings)
        return scores

    @property
    def __name__(self):
        return {
            SimilarityFunction.dot_product: dot_score,
            SimilarityFunction.cos_sim: cos_sim,
        }[self].__name__


@dataclass
class RetrievalTrainingExample:
    query: Query
    passages: List[Passage]  # (1 positive + n negatives)


class Separator(str, Enum):
    bert_sep = "bert_sep"
    roberta_sep = "roberta_sep"
    blank = "blank"
    empty = "empty"

    @property
    def token(self) -> str:
        return {
            Separator.bert_sep: "[SEP]",
            Separator.roberta_sep: "</s>",
            Separator.blank: " ",
            Separator.empty: "",
        }[self]

    def concat(self, texts: Iterator[Optional[str]]) -> str:
        """Concatenate two pieces of texts with the separation symbol."""
        return self.token.join(filter(lambda text: text is not None, texts))


@dataclass
class RetrieverConfig:
    query_model_name_or_path: str
    shared_encoder: bool
    sep: Separator
    pooling: Pooling
    similarity_function: SimilarityFunction
    max_length: int
    passage_model_name_or_path: Optional[str] = None
    sim_scale: float = 1.0  # For training, used to enlarge gradient

    def __post_init__(self) -> None:
        self.sep = Separator(self.sep)
        self.pooling = Pooling(self.pooling)
        self.similarity_function = SimilarityFunction(self.similarity_function)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def to_json_string(self) -> str:
        return json.dumps(self.to_dict())


class Retriever(torch.nn.Module):
    CONFIG_FNAME = "retriever_config.json"
    SEPARATE_PASSAGE_ENCODER = "separate_passage_encoder"
    PROMPT_SPACE_TOKEN = "[SPACE]"

    _keys_to_ignore_on_save = None  # Required by HF's trainer

    def __init__(self, config: RetrieverConfig) -> None:
        super().__init__()
        self.config = config
        self.tokenizer = AutoTokenizer.from_pretrained(config.query_model_name_or_path)
        self.query_encoder: PreTrainedModel = AutoModelForTextEncoding.from_pretrained(
            config.query_model_name_or_path
        )
        self.passage_encoder = self.query_encoder
        if not config.shared_encoder:
            assert config.passage_model_name_or_path is not None
            self.passage_encoder: PreTrainedModel = (
                AutoModelForTextEncoding.from_pretrained(
                    config.passage_model_name_or_path
                )
            )
        self.sep = Separator(config.sep)
        self.pooling = Pooling(config.pooling)
        self.similarity_function = SimilarityFunction(config.similarity_function)
        self.max_length = config.max_length
        self.set_device("cuda" if torch.cuda.is_available() else "cpu")
        self.query_prompt: Optional[str] = None
        self.passage_prompt: Optional[str] = None

    def set_query_prompt(self, prompt: str) -> None:
        self.query_prompt = prompt.replace(self.PROMPT_SPACE_TOKEN, " ")
        logging.info(f"Set query prompot to {self.query_prompt}")

    def set_passage_prompt(self, prompt: str) -> None:
        self.passage_prompt = prompt.replace(self.PROMPT_SPACE_TOKEN, " ")
        logging.info(f"Set passage prompot to {self.passage_prompt}")

    @staticmethod
    def maybe_add_prompt(prompt: Optional[str], input_text: str) -> str:
        if prompt is not None:
            return prompt + input_text
        else:
            return input_text

    def set_device(self, device: Union[int, str]) -> None:
        self.device = torch.device(device)
        self.to(self.device)

    def encode(
        self, encoder: PreTrainedModel, texts: List[str], batch_size: int
    ) -> torch.Tensor:
        encoded_list: List[torch.Tensor] = []
        for b in range(0, len(texts), batch_size):
            e = b + batch_size
            tokenized: BatchEncoding = self.tokenizer(
                texts[b:e],
                padding=True,
                truncation="longest_first",
                return_tensors="pt",
                max_length=self.max_length,
            ).to(self.device)
            outputs: BaseModelOutputWithPoolingAndCrossAttentions = encoder(**tokenized)
            pooled = self.pooling(
                token_embeddings=outputs.last_hidden_state,
                attention_mask=tokenized["attention_mask"],
            )  # (bsz, hdim)
            encoded_list.append(pooled)
        return torch.cat(encoded_list)

    def encode_queries(
        self, queries: List[Query], batch_size: int = 16
    ) -> torch.Tensor:
        return self.encode(
            encoder=self.query_encoder,
            texts=[
                self.maybe_add_prompt(prompt=self.query_prompt, input_text=query.text)
                for query in queries
            ],
            batch_size=batch_size,
        )

    def encode_passages(
        self, passages: List[Passage], batch_size: int = 16
    ) -> torch.Tensor:
        return self.encode(
            encoder=self.passage_encoder,
            texts=[
                self.maybe_add_prompt(
                    prompt=self.passage_prompt,
                    input_text=self.sep.concat([passage.title, passage.text]),
                )
                for passage in passages
            ],
            batch_size=batch_size,
        )

    def forward(self, examples: List[RetrievalTrainingExample]) -> torch.Tensor:
        queries = [e.query for e in examples]
        passages = list(chain(*(e.passages for e in examples)))
        qembs = self.encode_queries(queries)  # (bsz, hdim)
        pembs = self.encode_passages(passages)  # (bsz * (1 + nnegs), hdim)
        if dist.is_initialized():
            # Key in multi-gpu training for contrastive learning:
            # https://amsword.medium.com/gradient-backpropagation-with-torch-distributed-all-gather-9f3941a381f8
            qembs = dist_gather_tensor(qembs)
            pembs = dist_gather_tensor(pembs)
        sim_mtrx = self.similarity_function(qembs, pembs) * self.config.sim_scale
        npsgs_per_query = len(pembs) // len(qembs)
        labels = (
            torch.arange(sim_mtrx.size(0), device=sim_mtrx.device, dtype=torch.long)
            * npsgs_per_query
        )
        loss = torch.nn.functional.cross_entropy(sim_mtrx, labels)
        if dist.is_initialized():
            loss *= dist.get_world_size()
        return {"loss": loss}  # Align with HF's Trainer

    def save(self, output_dir) -> None:
        self.tokenizer.save_pretrained(output_dir)
        self.query_encoder.save_pretrained(output_dir)
        if not self.config.shared_encoder:
            self.passage_encoder.save_pretrained(
                os.path.join(output_dir, self.SEPARATE_PASSAGE_ENCODER)
            )
        config_dict = dataclasses.asdict(self.config)
        with open(os.path.join(output_dir, self.CONFIG_FNAME), "w") as f:
            json.dump(config_dict, f, indent=4)
        logging.info(f"Saved to {output_dir}")

    @classmethod
    def from_pretrained(cls: Type[Retriever], checkpoint_dir: str) -> Retriever:
        with open(os.path.join(checkpoint_dir, cls.CONFIG_FNAME)) as f:
            config = RetrieverConfig(**json.load(f))
        retriever = cls(config)
        retriever.query_encoder = AutoModelForTextEncoding.from_pretrained(
            checkpoint_dir
        )
        retriever.passage_encoder = retriever.query_encoder
        if not config.shared_encoder:
            retriever.passage_encoder = AutoModelForTextEncoding.from_pretrained(
                os.path.join(checkpoint_dir, cls.SEPARATE_PASSAGE_ENCODER)
            )
        retriever.set_device("cuda" if torch.cuda.is_available() else "cpu")
        logging.info(f"Loaded from {checkpoint_dir}")
        return retriever


@dataclass
class RetrieverBuildingArguments(RetrieverConfig):
    output_dir: Optional[str] = None


if __name__ == "__main__":
    # Build/convert a retriever into the supported format
    from clddp.utils import parse_cli, set_logger_format

    set_logger_format()
    args = parse_cli(RetrieverBuildingArguments)
    args_dict = dict(vars(args))
    args_dict.pop("output_dir")
    Retriever(RetrieverConfig(**args_dict)).save(args.output_dir)
    logging.info(f"Saved retriever checkpoint to {args.output_dir}")



from __future__ import annotations
import json
import logging
import os
import shutil
from typing import Any, Dict, List, Optional
import wandb
import torch
import torch.distributed as dist
from transformers import Trainer
from torch.utils.data import Dataset
from clddp.dm import JudgedPassage, RetrievalDataset, RetrievedPassageIDList, Split
from clddp.evaluation import search_and_evaluate
from clddp.args.train import RetrievalTrainingArguments
from clddp.retriever import RetrievalTrainingExample, Retriever, RetrieverConfig
from clddp.utils import is_device_zero, set_logger_format
from clddp.dataloader import load_dataset


class RetrievalTrainer(Trainer):
    @property
    def retriever(self) -> Retriever:
        return self.model

    def _save(self, output_dir: Optional[str] = None) -> None:
        self.retriever.save(output_dir)

    def training_step(
        self, model: torch.nn.Module, inputs: Dict[str, torch.Tensor | Any]
    ) -> torch.Tensor:
        loss = super().training_step(model, inputs)
        if dist.is_initialized():
            loss /= dist.get_world_size()  # Scale back the value for logging correctly
        return loss

    def evaluate(
        self,
        eval_dataset: Optional[RetrievalDataset] = None,
        ignore_keys: Optional[List[str]] = None,
        metric_key_prefix: str = "eval",
    ) -> Optional[Dict[str, float]]:
        args: RetrievalTrainingArguments = self.args
        if args.do_dev:
            dev_dataset = self.eval_dataset
            assert dev_dataset is not None
        else:
            return None
        _, report = search_and_evaluate(
            retriever=self.retriever,
            eval_dataset=dev_dataset,
            split=args.dev_split,
            topk=args.topk,
            per_device_eval_batch_size=args.per_device_eval_batch_size,
            fp16=args.fp16,
            metric_key_prefix=metric_key_prefix,
        )
        self.control.should_evaluate = (
            False  # Otherwise it will evaluate it again for epoch ends
        )
        if is_device_zero():
            self.log(report)
            return report
        else:
            return None


class RetrievalTrainingData(Dataset):
    def __init__(self, training_data: List[JudgedPassage]):
        self.training_data = training_data

    def __getitem__(self, item: int):
        jp = self.training_data[item]
        return RetrievalTrainingExample(query=jp.query, passages=[jp.passage])

    def __len__(self):
        return len(self.training_data)


def run(
    retriever: Retriever,
    args: RetrievalTrainingArguments,
    training_data: Optional[RetrievalTrainingData] = None,
    dev_dataset: Optional[RetrievalDataset] = None,
    test_dataset: Optional[RetrievalDataset] = None,
) -> None:
    if is_device_zero():
        args.dump_arguments()

    if args.do_dev:
        assert dev_dataset is not None
    if args.do_train:
        assert training_data is not None
        # Begin training:
        trainer = RetrievalTrainer(
            model=retriever,
            args=args,
            data_collator=lambda examples: {"examples": examples},
            train_dataset=training_data,
            eval_dataset=dev_dataset,
        )
        trainer.control.should_evaluate = True
        trainer._maybe_log_save_evaluate(
            tr_loss=None,
            model=retriever,
            trial=None,
            epoch=0,
            ignore_keys_for_eval=None,
        )
        trainer.train()
        logging.info("done training")

        # Saving:
        if is_device_zero():
            if trainer.state.best_model_checkpoint:
                shutil.copytree(
                    trainer.state.best_model_checkpoint,
                    os.path.dirname(
                        os.path.abspath(trainer.state.best_model_checkpoint)
                    ),
                    dirs_exist_ok=True,
                )
                logging.info(
                    f"Saved the best checkpoint from {trainer.state.best_model_checkpoint}"
                )
            else:
                trainer.save_model()
                logging.info("Saved the last checkpoint")

        # Make sure on the same page before evaluation:
        if dist.is_initialized():
            dist.barrier()

    # Final evaluation:
    if args.do_test:
        if os.path.exists(args.output_dir):
            retriever = Retriever.from_pretrained(args.output_dir)
            logging.info(f"Loaded from {args.output_dir} for evaluation")
            if args.query_prompt is not None:
                retriever.set_query_prompt(args.query_prompt)
            if args.passage_prompt is not None:
                retriever.set_passage_prompt(args.passage_prompt)
        else:
            assert not args.do_train, "No checkpoints saved after training"
        retrieval_results, report = search_and_evaluate(
            retriever=retriever,
            eval_dataset=test_dataset,
            split=args.test_split,
            topk=args.topk,
            fp16=args.fp16,
            per_device_eval_batch_size=args.per_device_eval_batch_size,
        )
        if is_device_zero():
            RetrievedPassageIDList.dump_trec_csv(
                retrieval_results=retrieval_results,
                fpath=os.path.join(args.output_dir, "retrieval_results.txt"),
            )
            freport = os.path.join(args.output_dir, "metrics.json")
            with open(freport, "w") as f:
                json.dump(report, f, indent=4)
            logging.info(f"Saved evaluation metrics to {freport}.")
            if wandb.run is None:
                wandb.init(
                    project=args.project,
                    name=args.run_name,
                    id=os.environ.get("WANDB_RUN_ID", None),
                )
            wandb_summary: dict = wandb.summary
            wandb_summary.update(report)
            wandb.finish()


if __name__ == "__main__":
    # Example cli: torchrun --nproc_per_node=4 --master_port=29501 -m clddp.train
    from clddp.utils import parse_cli

    set_logger_format(logging.INFO if is_device_zero() else logging.WARNING)
    args = parse_cli(RetrievalTrainingArguments)

    # Retriever building:
    config = RetrieverConfig(
        query_model_name_or_path=args.query_model_name_or_path,
        passage_model_name_or_path=args.passage_model_name_or_path,
        shared_encoder=args.shared_encoder,
        sep=args.sep,
        pooling=args.pooling,
        similarity_function=args.similarity_function,
        max_length=args.max_length,
        sim_scale=args.sim_scale,
    )
    retriever = Retriever(config)
    if args.query_prompt is not None:
        retriever.set_query_prompt(args.query_prompt)
    if args.passage_prompt is not None:
        retriever.set_passage_prompt(args.passage_prompt)

    # Data loading:
    train_dataset = load_dataset(
        enable=args.do_train,
        dataloader_name=args.train_dataloader,
        data_name_or_path=args.train_data,
    )
    training_data = RetrievalTrainingData(train_dataset.judged_passages_train)
    dev_dataset = train_dataset
    if args.dev_data != args.train_data:
        dev_dataset = load_dataset(
            enable=args.do_dev,
            dataloader_name=args.dev_dataloader,
            data_name_or_path=args.dev_data,
        )
    if args.quick_dev:
        assert args.do_dev
        save_pids_to_fpath = (
            os.path.join(args.output_dir, "quick_dev_pids.json")
            if is_device_zero()
            else None
        )
        dev_dataset = dev_dataset.to_quick_version(
            split=Split.dev,
            progress_bar=is_device_zero(),
            save_pids_to_fpath=save_pids_to_fpath,
        )
    test_dataset = train_dataset
    if args.test_data != args.train_data:
        test_dataset = load_dataset(
            enable=args.do_test,
            dataloader_name=args.test_dataloader,
            data_name_or_path=args.test_data,
        )
    if args.quick_test:
        assert args.do_test
        save_pids_to_fpath = (
            os.path.join(args.output_dir, "quick_test_pids.json")
            if is_device_zero()
            else None
        )
        test_dataset = test_dataset.to_quick_version(
            split=Split.test,
            progress_bar=is_device_zero(),
            save_pids_to_fpath=save_pids_to_fpath,
        )

    # Run training:
    run(
        retriever=retriever,
        args=args,
        training_data=training_data,
        dev_dataset=dev_dataset,
        test_dataset=test_dataset,
    )
    logging.info("Done")



from __future__ import annotations
from dataclasses import asdict, dataclass
import json
import logging
import os
from typing import Optional, Set, Type
from clddp.utils import get_commit_hash, parse_cli
from abc import ABC, abstractmethod


@dataclass
class AutoRunNameArgumentsMixIn:
    run_name: Optional[str] = None

    @property
    def escaped_args(self) -> Set[str]:
        """Escaped items for building the run_name."""
        return set()

    @property
    def auto_run_name(self) -> str:
        return "/".join(
            [
                f"{k}_{getattr(self, k)}"
                for k in self.__annotations__.keys()
                if k not in self.escaped_args
            ]
            + [f"git_hash_{get_commit_hash()}"]
        )

    def __post_init__(self) -> None:
        if self.run_name is None:
            self.run_name = self.auto_run_name
        if hasattr(super(), "__post_init__"):
            super().__post_init__()  # Needed as a MixIn


@dataclass
class DumpableArgumentsMixIn(ABC):
    output_dir: Optional[str] = None

    def __post_init__(self) -> None:
        if self.output_dir is None:
            self.output_dir = self.build_output_dir()
        if hasattr(super(), "__post_init__"):
            super().__post_init__()  # Needed as a MixIn

    @abstractmethod
    def build_output_dir(self) -> str:
        pass

    def dump_arguments(self, fname: str = "arguments.json") -> None:
        os.makedirs(self.output_dir, exist_ok=True)
        fargs = os.path.join(self.output_dir, fname)
        with open(fargs, "w") as f:
            json.dump(asdict(self), f, indent=4)
        logging.info(f"Dumped arguments to {fargs}")



import logging
import os
import pickle
from typing import Iterable, Iterator, List, Type, TypeVar
import git
import numpy as np
import torch.distributed as dist
import torch
import tqdm
from transformers import HfArgumentParser


NINF = -1e4


def is_device_zero() -> bool:
    return os.environ.get("LOCAL_RANK", "0") == "0"


def set_logger_format() -> None:
    logging.basicConfig(level=logging.INFO)
    root_logger = logging.getLogger()
    formatter = logging.Formatter(
        fmt="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    for handler in root_logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.setFormatter(formatter)


OBJ_TYPE = TypeVar("OBJ_TYPE")


def broadcast_object(obj: OBJ_TYPE) -> OBJ_TYPE:
    """Broadcast a Python object across all the nodes."""
    # Collect object from replicas and form a list
    if not dist.is_initialized():
        return obj

    rank = dist.get_rank()
    obj_size: torch.LongTensor = torch.zeros(1).long().to(rank)  # long not int!
    if rank == 0:
        data = pickle.dumps(obj)
        data_length = len(data)
        data = data_length.to_bytes(4, "big") + data
        data = np.frombuffer(data, dtype=np.uint8)
        obj_size += len(data)
        tensorized: torch.Tensor = torch.from_numpy(data).to(rank)
        logging.info(f"Going to broacast {obj_size.item() / 2**20:.1f}MB")
    dist.broadcast(obj_size, 0)
    if rank != 0:
        tensorized = torch.zeros(obj_size.item(), dtype=torch.uint8).to(rank)
    dist.broadcast(tensorized, 0)
    tensorized_numpy: np.ndarray = tensorized.cpu().numpy()
    data = tensorized_numpy.tobytes()
    del tensorized
    torch.cuda.empty_cache()
    length = int.from_bytes(data[:4], "big")
    data = data[4 : length + 4]
    obj: OBJ_TYPE = pickle.loads(data)
    return obj


def all_gather_object(obj: OBJ_TYPE) -> List[OBJ_TYPE]:
    """Broadcast a Python object across all the nodes."""
    if not dist.is_initialized():
        return [obj]

    # Collect object from replicas and form a list
    ngpus = dist.get_world_size()
    obj_list = [None for _ in range(ngpus)]
    dist.all_gather_object(obj_list, obj)
    return obj_list


def split_data(data: Iterable[OBJ_TYPE]) -> Iterable[OBJ_TYPE]:
    if not dist.is_initialized():
        for datum in data:
            yield datum

    rank = dist.get_rank()
    ngpus = dist.get_world_size()
    for i, datum in enumerate(data):
        if i % ngpus == rank:
            yield datum


def split_data_size(data_size: int) -> int:
    if not dist.is_initialized():
        return data_size

    rank = dist.get_rank()
    ngpus = dist.get_world_size()
    return data_size // ngpus + int(rank < data_size % ngpus)


def dist_gather_tensor(t: torch.Tensor) -> torch.Tensor:
    assert dist.is_initialized()
    ngpus = dist.get_world_size()
    rank = dist.get_rank()

    if t is None:
        return None
    t = t.contiguous()

    all_tensors = [torch.empty_like(t) for _ in range(ngpus)]
    dist.all_gather(all_tensors, t)

    all_tensors[rank] = t
    all_tensors = torch.cat(all_tensors, dim=0)

    return all_tensors


def tqdm_ropen(fpath: str, desc: str, pbar: bool) -> Iterator[str]:
    """tqdm + open with r mode."""
    if desc is None:
        desc = f"Loading from {fpath}"

    with open(fpath, "r") as f:
        nlines = sum(1 for _ in f)

    with tqdm.tqdm(open(fpath, "r"), desc=desc, total=nlines, disable=not pbar) as f:
        for line in f:
            yield line


def set_logger_format(level: int = logging.INFO) -> None:
    logging.basicConfig(level=level)
    root_logger = logging.getLogger()
    formatter = logging.Formatter(
        fmt="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    for handler in root_logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.setFormatter(formatter)


def get_rank() -> int:
    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def get_ngpus() -> int:
    if not dist.is_initialized():
        return int(torch.cuda.is_available())
    return dist.get_world_size()


def initialize_ddp(ddp_timeout: int = 360000) -> None:
    from transformers import TrainingArguments

    TrainingArguments(
        "dummy", ddp_timeout=ddp_timeout
    )  # This will invoke dist.init_process_group
    cuda_env = os.environ.get("CUDA_VISIBLE_DEVICES")
    if cuda_env:
        os.environ["CUDA_VISIBLE_DEVICES"] = cuda_env.split(",")[get_rank()]
    assert dist.is_initialized()


def get_commit_hash() -> str:
    """Return the HEAD commit hash."""
    repo = git.Repo(search_parent_directories=True)
    sha = repo.head.object.hexsha
    return sha


def parse_cli(arguments_class: Type[OBJ_TYPE]) -> OBJ_TYPE:
    parser = HfArgumentParser(arguments_class)
    args = arguments_class(**vars(parser.parse_args()))
    return args


def colbert_score_reduce(
    scores_padded: torch.Tensor,
    D_mask: torch.Tensor,
    interaction: str = "colbert",
    query_maxlen: int = 64,
):
    """Copied and modified from ColBERT."""
    D_padding = ~D_mask.view(scores_padded.size(0), scores_padded.size(1)).bool()
    scores_padded[D_padding] = -9999
    scores: torch.Tensor = scores_padded.max(1).values
    assert interaction in ["colbert", "flipr"]
    if interaction == "flipr":
        K1 = query_maxlen // 2
        K2 = 8
        A = scores[:, :query_maxlen].topk(K1, dim=-1).values.sum(-1)
        B = 0
        if K2 <= scores.size(1) - query_maxlen:
            B = scores[:, query_maxlen:].topk(K2, dim=-1).values.sum(1)
        return A + B
    return scores.sum(-1)


def colbert_score(Q: torch.Tensor, D_padded: torch.Tensor, D_mask: torch.BoolTensor):
    """
    Copied and modified from ColBERT.
    Supply sizes Q = (1 | num_docs, *, dim) and D = (num_docs, *, dim).
    If Q.size(0) is 1, the matrix will be compared with all passages.
    Otherwise, each query matrix will be compared against the *aligned* passage.

    EVENTUALLY: Consider masking with -inf for the maxsim (or enforcing a ReLU).
    """

    assert Q.dim() == 3, Q.size()
    assert D_padded.dim() == 3, D_padded.size()
    assert Q.size(0) in [1, D_padded.size(0)]
    scores = D_padded @ Q.to(dtype=D_padded.dtype).permute(0, 2, 1)
    return colbert_score_reduce(scores, D_mask)



from __future__ import annotations
from dataclasses import dataclass
import logging
import os
from typing import Dict, Iterable, List, Optional, Type
from more_itertools import chunked
import torch
import tqdm
from clddp.dm import (
    JudgedPassage,
    Passage,
    Query,
    RetrievedPassageIDList,
    ScoredPassageID,
)
from clddp.args.search import SearchArguments
from clddp.dataloader import load_dataset
from clddp.utils import (
    NINF,
    get_rank,
    all_gather_object,
    is_device_zero,
    set_logger_format,
    split_data,
    split_data_size,
)
from clddp.retriever import Retriever


@dataclass
class SearchOutput:
    wid: int  # worker index
    topk_indices: torch.LongTensor  # (nqueries, topk)
    topk_values: torch.Tensor  # (nqueries, topk)

    def update(
        self,
        base: Optional[int],
        similarity_matrix: torch.Tensor,
        nqueries: int,
        topk: int,
    ) -> int:
        nentries: int = similarity_matrix.shape[1]  # either #passages or #documents
        assert similarity_matrix.shape[0] == nqueries
        self.topk_values, topk_setoffs = torch.cat(
            [self.topk_values, similarity_matrix], dim=-1
        ).topk(k=topk, dim=1, largest=True, sorted=False)
        new_indices = (
            torch.arange(nentries)
            .to(self.topk_indices)
            .unsqueeze(0)
            .expand(nqueries, nentries)
        ) + base
        self.topk_indices = torch.cat([self.topk_indices, new_indices], dim=-1).gather(
            dim=1, index=topk_setoffs
        )
        return nentries

    @classmethod
    def merge(
        cls: Type[SearchOutput], souts: List[SearchOutput], topk: int
    ) -> SearchOutput:
        assert len(souts)
        topk_values, topk_setoffs = torch.cat(
            [sout.topk_values for sout in souts], dim=-1
        ).topk(k=topk, dim=1, largest=True, sorted=False)
        topk_indices = torch.cat([sout.topk_indices for sout in souts], dim=-1).gather(
            dim=1, index=topk_setoffs
        )
        return SearchOutput(
            wid=-1,
            topk_indices=topk_indices,
            topk_values=topk_values,
        )

    def to_retrieved_passage_id_lists(
        self, queries: List[Query], passage_ids: List[str], pbar: bool
    ) -> List[RetrievedPassageIDList]:
        topk_indices = self.topk_indices.cpu().tolist()
        topk_values = self.topk_values.cpu().tolist()
        retrieval_results: List[RetrievedPassageIDList] = []
        for i, query in enumerate(
            tqdm.tqdm(queries, desc="Formatting results", disable=not pbar)
        ):
            spids: List[ScoredPassageID] = []
            pids = map(
                lambda topk_index: passage_ids[topk_index],
                topk_indices[i],
            )
            for pid, score in zip(pids, topk_values[i]):
                spids.append(ScoredPassageID(passage_id=pid, score=score))
            retrieval_results.append(
                RetrievedPassageIDList(
                    query_id=query.query_id, scored_passage_ids=spids
                )
            )
        return retrieval_results

    def cpu(self) -> SearchOutput:
        return SearchOutput(
            wid=self.wid,
            topk_indices=self.topk_indices.cpu(),
            topk_values=self.topk_values.cpu(),
        )


@torch.no_grad()
def search_single_device(
    retriever: Retriever,
    queries: List[Query],
    collection_iter: Iterable[Passage],
    collection_size: int,
    topk: int,
    batch_size: int,
    fp16: bool,
    wid: int = 0,
    show_pbar: bool = True,
) -> List[RetrievedPassageIDList]:
    retriever.eval()
    qembs = retriever.encode_queries(queries=queries, batch_size=batch_size)
    search_output = SearchOutput(
        wid=wid,
        topk_indices=torch.zeros(len(qembs), topk, dtype=torch.int64).to(
            retriever.device
        ),
        topk_values=torch.full((len(qembs), topk), NINF).to(retriever.device),
    )
    batches = chunked(iterable=collection_iter, n=batch_size)
    pids: List[str] = []
    pbar = tqdm.tqdm(total=collection_size, desc="Searching", disable=not show_pbar)
    for batch in batches:
        with torch.cuda.amp.autocast(enabled=fp16):
            pembs = retriever.encode_passages(passages=batch, batch_size=batch_size)
        sim_mtrx = retriever.similarity_function(
            query_embeddings=qembs, passage_embeddings=pembs
        )
        nentires = search_output.update(
            base=len(pids),
            similarity_matrix=sim_mtrx,
            nqueries=len(qembs),
            topk=topk,
        )
        pids.extend(psg.passage_id for psg in batch)
        pbar.update(nentires)
    return search_output.to_retrieved_passage_id_lists(
        queries=queries, passage_ids=pids, pbar=show_pbar
    )


def search(
    retriever: Retriever,
    collection_iter: Iterable[Passage],
    collection_size: int,
    queries: List[Query],
    topk: int,
    per_device_eval_batch_size: int,
    fp16: bool,
) -> List[RetrievedPassageIDList]:
    # Doing search:
    rank = get_rank()
    split_collection_iter = split_data(data=collection_iter)
    split_collection_size = split_data_size(collection_size)
    retrieved = search_single_device(
        retriever=retriever,
        queries=queries,
        collection_iter=split_collection_iter,
        collection_size=split_collection_size,
        topk=topk,
        batch_size=per_device_eval_batch_size,
        fp16=fp16,
        wid=rank,
        show_pbar=is_device_zero(),
    )
    all_retrieved = sum(all_gather_object(retrieved), [])
    qid2rpidls: Dict[str, List[RetrievedPassageIDList]] = {}
    for rpidl in all_retrieved:
        qid2rpidls.setdefault(rpidl.query_id, [])
        qid2rpidls[rpidl.query_id].append(rpidl)
    merged_rpidls = []
    for rpidls in tqdm.tqdm(
        list(qid2rpidls.values()),
        desc="Merging search results",
        disable=not is_device_zero(),
    ):
        rpidl = RetrievedPassageIDList.merge(rpidls=rpidls, topk=topk)
        merged_rpidls.append(rpidl)
    return merged_rpidls


if __name__ == "__main__":
    from clddp.utils import parse_cli, initialize_ddp

    set_logger_format(logging.INFO if is_device_zero() else logging.WARNING)
    initialize_ddp()
    args = parse_cli(SearchArguments)
    if is_device_zero():
        args.dump_arguments()

    retriever = Retriever.from_pretrained(args.checkpoint_dir)
    if args.query_prompt is not None:
        retriever.set_query_prompt(args.query_prompt)
    if args.passage_prompt is not None:
        retriever.set_passage_prompt(args.passage_prompt)
    retriever.set_device(get_rank())
    dataset = load_dataset(
        enable=True, dataloader_name=args.dataloader, data_name_or_path=args.data_dir
    )
    jpsgs = dataset.get_judged_passages(args.split)
    queries = JudgedPassage.get_unique_queries(jpsgs)
    retrieval_results = search(
        retriever=retriever,
        collection_iter=dataset.collection_iter,
        collection_size=dataset.collection_size,
        queries=queries,
        topk=args.topk,
        per_device_eval_batch_size=args.per_device_eval_batch_size,
        fp16=args.fp16,
    )
    if is_device_zero():
        fretrieved = os.path.join(args.output_dir, "retrieval_results.txt")
        RetrievedPassageIDList.dump_trec_csv(
            retrieval_results=retrieval_results, fpath=fretrieved
        )
    logging.info("Done")
